// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "ShaderProgram.hpp"
#include "assert/assert.hpp"
#include "assert/assertgl.hpp"

#include "component/display/material/SimpleTexturedMaterial.hpp"

using namespace Dawn;

struct TestData {
  glm::mat4 projection;
  glm::mat4 view;
  glm::mat4 model;
  int hasTexture;
  int padding[3];
  struct Color color;
  int test;
};

void ShaderProgram::init(
  IModule *module,
  Slang::ComPtr<ISession> session
) {
  assertTrue(this->id == -1, "ShaderProgram already initialized?");
  IShaderProgram::init(module, session);

  // Create the program
  this->id = glCreateProgram();
  assertNoGLError();

  // Attach all the shader stages
  for(auto stage : stages) {
    glAttachShader(this->id, stage->id);
    assertNoGLError();
  }

  // Link and verify the program
  glLinkProgram(this->id);
  assertNoGLError();

  GLint status;
  glGetProgramiv(this->id, GL_LINK_STATUS, &status);
  assertNoGLError();

  if(!status) {
    // Failed to link
    GLint logLength;
    glGetProgramiv(this->id, GL_INFO_LOG_LENGTH, &logLength);
    assertNoGLError();

    GLchar *log = new GLchar[logLength];
    glGetProgramInfoLog(this->id, logLength, NULL, log);
    assertNoGLError();
    assertUnreachable("Failed to link shader program:\n%s", log);
  }

  // So the uniforms that are in slang are kinda odd when compiled.

  //DEBUGGING
  // GLint numUniforms = 0;
  // glGetProgramiv(this->id, GL_ACTIVE_UNIFORMS, &numUniforms);
  // assertNoGLError();
  // for(GLint i = 0; i < numUniforms; ++i) {
  //   char name[1024];
  //   GLsizei length;
  //   GLint size;
  //   GLenum type;
  //   glGetActiveUniform(this->id, i, sizeof(name), &length, &size, &type, name);
  //   assertNoGLError();
  //   std::cout << "Uniform: " << i << ": " << name << " has size " << size << " and length " << length << std::endl;
  // }
  // GLint numUniformBlocks = 0;
  // glGetProgramiv(this->id, GL_ACTIVE_UNIFORM_BLOCKS, &numUniformBlocks);
  // for(GLint i = 0; i < numUniformBlocks; ++i) {
  //   GLint size;
  //   glGetActiveUniformBlockiv(this->id, i, GL_UNIFORM_BLOCK_DATA_SIZE, &size);
  //   assertNoGLError();
  //   std::cout << "Uniform Block: " << i << " has size " << size << std::endl;
  // }
  // auto data = std::make_shared<SimpleTexturedMaterialShaderData>();
  // data->color = COLOR_WHITE;
  // data->model = glm::mat4(1.0f);
  // data->projection = glm::perspective(
  //   glm::radians(45.0f), 800.0f / 600.0f, 0.1f, 100.0f
  // );
  // data->view = glm::lookAt(
  //   glm::vec3(4,3,3), glm::vec3(0,0,0), glm::vec3(0,1,0)
  // );

  glUseProgram(this->id);
  assertNoGLError();
  GLuint ubo;
  glGenBuffers(1, &ubo);
  assertNoGLError();

  struct TestData data;
  data.color = COLOR_WHITE;
  data.model = glm::mat4(1.0f);
  data.projection = glm::perspective(
    glm::radians(45.0f), 800.0f / 600.0f, 0.1f, 10000.0f
  );
  data.view = glm::lookAt(
    glm::vec3(300,300,300), glm::vec3(0,0,0), glm::vec3(0,1,0)
  );
  data.hasTexture = false;
  data.test = 1;

  glBindBuffer(GL_UNIFORM_BUFFER, ubo);
  assertNoGLError();
  glBindBufferBase(GL_UNIFORM_BUFFER, 0, ubo);
  assertNoGLError();
  glBufferData(GL_UNIFORM_BUFFER, sizeof(struct TestData), &data, GL_STATIC_DRAW);
  assertNoGLError();

  // testTexture = std::make_shared<Texture>();
  // testTexture->setSize(2, 2, TextureFormat::RGBA, TextureDataFormat::UNSIGNED_BYTE);
  // uint8_t colors[4 * 4];
  // colors[0] = 255; colors[1] = 0; colors[2] = 0; colors[3] = 255;
  // colors[4] = 0; colors[5] = 255; colors[6] = 0; colors[7] = 255;
  // colors[8] = 0; colors[9] = 0; colors[10] = 255; colors[11] = 255;
  // colors[12] = 255; colors[13] = 255; colors[14] = 255; colors[15] = 255;
  // testTexture->buffer(colors);
  // testTexture->bind(0);

  // int index = glGetUniformLocation(this->id, "globalParams_uniforms_texture_0");
  // assertNoGLError();
  // std::cout << "Index: " << index << std::endl;
  // glUniform1i(index, 0);
  // assertNoGLError();  

  // data->write(shared_from_this());
}

ShaderProgram::~ShaderProgram() {
  // Delete the shader program
  if(this->id != -1) {
    glDeleteProgram(this->id);
    assertNoGLError();
  }
}